//*****************************************************
//**   Ivory Draw 1.0 Beta 4                         **
//**   DirectDraw component package for C++Builder   **
//*****************************************************

// (C) 1998 Tamas Demjen
// email: demjen@usa.net
// Web: http://members.xoom.com/demjen
// Ivory Draw on the Web: http://members.xoom.com/demjen/bcb/ivory.html

// This product is freeware, as described in license.txt
//---------------------------------------------------------------------------
#ifndef CustomIvorySurfaceH
#define CustomIvorySurfaceH

//#include "IvoryError.h"
#include "IvoryResFileBase.h"
#include "IvoryCanvas.h"
//---------------------------------------------------------------------------
class IVORY_package TIvoryDraw;
//class IVORY_package TIvoryMapper;

enum TSurfaceType { stUnspecified, stPrimary, stBack, stOffScreen };

typedef void __fastcall (__closure* TRequestSizeEvent)(System::TObject* Sender,
                         int& Width, int& Height);

class IVORY_package TCustomIvorySurface : public TIvoryResFileBase
{
   friend TIvoryDraw;
private:
// internal handles
   LPDIRECTDRAWSURFACE FDirectDrawSurface;
   LPDIRECTDRAWCLIPPER FDirectDrawClipper;
// child objects
   TIvoryCanvas* FCanvas;
// initial settings
   bool FAutoCreate;
   TSurfaceType FSurfaceType;
   int FWidth;
   int FHeight;
   bool FTransparent;
   int FTransparentColor;
// current state
   bool FLocked;
   unsigned char* FPixels;
   int FPitch;
   bool FClearBackground;
   int FBackgroundColor;
   bool FSubPaint;
   bool FPainted; // it is painted in the current cycle
   bool FValid;
// internal messages
   Classes::TNotifyEvent FOnCreateInternal;
   Classes::TNotifyEvent FOnSurfaceLost;
   TRequestSizeEvent FOnRequestSize;
// public messages
   Classes::TNotifyEvent FOnPaint;
   Classes::TNotifyEvent FOnCreate;
   Classes::TNotifyEvent FOnDestroy;
   Classes::TNotifyEvent FOnChange;
// getter and setter functions
   unsigned char* __fastcall GetScanLine(int Row) const;
   void __fastcall SetAutoCreate(bool ac);
   void __fastcall SetWidth(int width);
   void __fastcall SetHeight(int height);
   void __fastcall SetTransparent(bool transparent);
   void __fastcall SetTransparentColor(int tcolor);
   void __fastcall SetClearBackground(bool cb);
   void __fastcall SetBackgroundColor(int bc);
   void __fastcall SetLocked(bool locked);
// internal functions
   void __fastcall CreatePrimary();
   void __fastcall CreateBack(TCustomIvorySurface* primary);
   void __fastcall InternalCreateOffScreen(); // FWidth and FHeight must be set!
   void __fastcall Release();
protected:
// internal functions
   inline void __fastcall Restore() { DirectDrawSurface->Restore(); }
   virtual void __fastcall RealizePaint();
   void __fastcall StartPainting();
   void __fastcall FinishPainting();
// internal properties
// internal events
   __property Classes::TNotifyEvent OnCreateInternal = { read=FOnCreateInternal, write=FOnCreateInternal };
   __property Classes::TNotifyEvent OnSurfaceLost = { read=FOnSurfaceLost, write=FOnSurfaceLost };
   __property TRequestSizeEvent OnRequestSize = { read=FOnRequestSize, write=FOnRequestSize };
public:
// constructor
   __fastcall TCustomIvorySurface(TComponent* AOwner);
// destructor
   virtual __fastcall ~TCustomIvorySurface();
// public functions
   void __fastcall Create();
   inline void __fastcall CreateOffScreen() { Create(); }
   void __fastcall Recreate(int width, int height);
   inline void __fastcall RecreateOffScreen(int width, int height) { Recreate(width, height); }
   void __fastcall Recreate(int width, int height, bool transparent, int color = 0);
   inline void __fastcall RecreateOffScreen(int width, int height, bool transparent, int color = 0)
      { Recreate(width, height, transparent, color); }
   void __fastcall Paint(bool subpaint = false);
   inline void __fastcall Invalidate() { Valid = false; }
   void __fastcall Lock();
   void __fastcall Unlock();
   void __fastcall GrabCanvas();
   void __fastcall ReleaseCanvas();
   void __fastcall CopySurface(int origx, int origy, TCustomIvorySurface* surface);
   void __fastcall CopySurface(int origx, int origy, TCustomIvorySurface* surface, const TRect rect);
   void __fastcall StretchCopySurface(const TRect destrect, TCustomIvorySurface* surface);
   void __fastcall StretchCopySurface(const TRect destrect, TCustomIvorySurface* surface, const TRect rect);
   void __fastcall TileCopySurface(const TRect destrect, TCustomIvorySurface* surface);
   void __fastcall TileCopySurface(const TRect destrect, TCustomIvorySurface* surface, const TRect rect);
   void __fastcall FillRect(const TRect rect, int color = 0);
   void __fastcall PaintSubSurface(TCustomIvorySurface* surface) { if(SubPaint) surface->Paint(true); }
// public properties
   __property LPDIRECTDRAWSURFACE DirectDrawSurface = { read=FDirectDrawSurface };
   __property LPDIRECTDRAWCLIPPER DirectDrawClipper = { read=FDirectDrawClipper };
   __property TIvoryCanvas* Canvas = { read=FCanvas };
   __property TSurfaceType SurfaceType = { read=FSurfaceType };
   __property unsigned char* Pixels = { read=FPixels };
   __property int Pitch = { read=FPitch };
   __property unsigned char* ScanLine[int Row] = { read=GetScanLine };
   __property bool Locked = { read=FLocked, write=SetLocked };
   __property bool SubPaint = { read=FSubPaint };
   __property bool Valid = { read=FValid, write=FValid };
// gonna be published properties in TIvorySurface
   __property bool AutoCreate = { read=FAutoCreate, write=SetAutoCreate, default=true };
   __property int Width = { read=FWidth, write=SetWidth, nodefault };
   __property int Height = { read=FHeight, write=SetHeight, nodefault };
   __property bool Transparent = { read=FTransparent, write=SetTransparent, default=false };
   __property int TransparentColor = { read=FTransparentColor, write=SetTransparentColor, default=0 };
   __property bool ClearBackground = { read=FClearBackground, write=SetClearBackground, default=true };
   __property int BackgroundColor = { read=FBackgroundColor, write=SetBackgroundColor, default=0 };
// gonna be published events in TIvorySurface
   __property Classes::TNotifyEvent OnPaint = { read=FOnPaint, write=FOnPaint, default=NULL };
   __property Classes::TNotifyEvent OnCreate = { read=FOnCreate, write=FOnCreate, default=NULL };
   __property Classes::TNotifyEvent OnDestroy = { read=FOnDestroy, write=FOnDestroy, default=NULL };
   __property Classes::TNotifyEvent OnChange = { read=FOnChange, write=FOnChange, default=NULL };
};
//---------------------------------------------------------------------------
#endif
