//*****************************************************
//**   Ivory Draw 1.0 Beta 4                         **
//**   DirectDraw component package for C++Builder   **
//*****************************************************

// (C) 1998 Tamas Demjen
// email: demjen@usa.net
// Web: http://members.xoom.com/demjen
// Ivory Draw on the Web: http://members.xoom.com/demjen/bcb/ivory.html

// This product is freeware, as described in license.txt
//---------------------------------------------------------------------------
#include <vcl.h>
#pragma hdrstop

#include "IvoryBitmapSurface.h"
#include "IvoryDraw.h"
#include "IvoryMiscUtils.h"
//---------------------------------------------------------------------------
#ifdef __IvoryBCB3__
#pragma package(smart_init)
#endif
//---------------------------------------------------------------------------
// internal variables:
static const TIvoryErrorClass iec = iecIvoryBitmapSurface;
//---------------------------------------------------------------------------
static inline void ValidCtrCheckBitmapSurface(TIvoryBitmapSurface*)
{
   new TIvoryBitmapSurface(NULL);
}
//---------------------------------------------------------------------------
__fastcall TIvoryBitmapSurface::TIvoryBitmapSurface(TComponent* AOwner)
 : TCustomIvorySurface(AOwner),
   Content(0), FOnPaintBitmap(NULL), FOnUserFileFormat(NULL)
{
   OnRequestSize = BitmapSurfaceRequestSize;
}
//---------------------------------------------------------------------------
__fastcall TIvoryBitmapSurface::~TIvoryBitmapSurface()
{
   if(Content == 1 && Memory) { delete [] Memory; Memory = NULL; Content = 0; }
}
//---------------------------------------------------------------------------
void __fastcall TIvoryBitmapSurface::RealizePaint()
{
   const char* err_msg = "Error loading image file";
   const char* ivoryfuncname = "RealizePaint";

   if(Content == 1)
   {
      bool need_lock_unlock = false;
      try
      {
         if(DirectDrawSurface == NULL) IVORYERROR(err_msg, ierIvorySurfaceNotCreated);
         if(ImageWidth != Width || ImageHeight != Height) IVORYERROR(err_msg, ierOther);
         if(FileType > 2)
         {
            if(OnPaintBitmap) OnPaintBitmap(this, Memory, Size, FileType);
         }
         else if(FileType == 1 && TheIvoryDraw->ExclusiveMode == false)
         {
            need_lock_unlock = false;
            LoadBMPFile(FileName);
         }
         else
         {
            need_lock_unlock = !Locked;
            if(need_lock_unlock) Lock();
            if(FileType == 1) FastLoadBMP(Memory, Width, Height, OffsetPixels, SizePixels);
            else if(FileType == 2) FastLoadPCX(Memory, Width, Height, OffsetPixels, SizePixels);
         }
      }
      catch(...)
      {
         if(need_lock_unlock) { Unlock(); need_lock_unlock = false; }
         if(Memory) { delete [] Memory; Memory = NULL; }
         Content = 0;
         throw;
      }
      if(need_lock_unlock) Unlock();
      if(Memory) delete [] Memory; Memory = NULL;
      Content = 3;
   }
   else if(Content == 3)
      LoadFromFile();
   else if(Content == 2)
   {
      bool need_lock_unlock = false;
      try
      {
         if(DirectDrawSurface == NULL) IVORYERROR(err_msg, ierIvorySurfaceNotCreated);
         if(ImageWidth != Width || ImageHeight != Height) IVORYERROR(err_msg, ierOther);
         if(FileType > 2)
         {
            if(OnPaintBitmap) OnPaintBitmap(this, Memory, Size, FileType);
         }
         else if(FileType == 1 && TheIvoryDraw->ExclusiveMode == false)
         {
            need_lock_unlock = false;
            LoadBMPResource(DllHandle, ResourceName);
         }
         else
         {
            need_lock_unlock = !Locked;
            if(need_lock_unlock) Lock();
            if(FileType == 1) FastLoadBMP(Memory, Width, Height, OffsetPixels, SizePixels);
            else if(FileType == 2) FastLoadPCX(Memory, Width, Height, OffsetPixels, SizePixels);
         }
      }
      catch(...)
      {
         if(need_lock_unlock) { Unlock(); need_lock_unlock = false; }
         Content = 0;
         throw;
      }
      if(need_lock_unlock) Unlock();
//      Content = 0;
   }
   else if(Content == 4)
      LoadFromResource();
}
//---------------------------------------------------------------------------
void __fastcall TIvoryBitmapSurface::BitmapSurfaceRequestSize(TObject* Sender, int& Width, int& Height)
{
   if(Content != 0)
   {
      Width = ImageWidth; Height = ImageHeight;
   }
   else
   {
      Width = 320; Height = 240;
   }
}
//---------------------------------------------------------------------------
void __fastcall TIvoryBitmapSurface::LoadBMPFile(const String filename)
{
   HBITMAP hbm;
   try
   {
      hbm = (HBITMAP)LoadImage(NULL, filename.c_str(), IMAGE_BITMAP, 0, 0, LR_LOADFROMFILE | LR_CREATEDIBSECTION);
      bool no_canvas = Canvas->Handle == NULL;
      if(no_canvas) GrabCanvas(); // get a device context
      Canvas->Draw(0,0, hbm); // draw the background bitmap stretched
      if(no_canvas) ReleaseCanvas(); // release the device context
   }
   catch(...)
   {
      DeleteObject(hbm); hbm = NULL;
      throw;
   }
   DeleteObject(hbm);
}
//---------------------------------------------------------------------------
void __fastcall TIvoryBitmapSurface::LoadBMPResource(HMODULE hModule, const String resname)
{
   HBITMAP hbm;
   try
   {
      if(hModule == NULL) hModule = GetModuleHandle(NULL);
      hbm = (HBITMAP)LoadImage(hModule, resname.c_str(), IMAGE_BITMAP, 0, 0, LR_CREATEDIBSECTION);
      bool no_canvas = Canvas->Handle == NULL;
      if(no_canvas) GrabCanvas(); // get a device context
      Canvas->Draw(0,0, hbm); // draw the background bitmap stretched
      if(no_canvas) ReleaseCanvas(); // release the device context
   }
   catch(...)
   {
      DeleteObject(hbm); hbm = NULL;
      throw;
   }
   DeleteObject(hbm);
}
//---------------------------------------------------------------------------
void __fastcall TIvoryBitmapSurface::FastLoadBMP(unsigned char* Memory, int width, int height, int offsetpixels, int sizepixels)
{
   try
   {
      Invalidate();
      if(OnChange) OnChange(this);

      Memory += offsetpixels;
      unsigned char* offscr = Pixels;
      Memory += width * (height - 1);
      if(TheIvoryDraw->Palette->OriginalColor[255] != clWhite ||
         TheIvoryDraw->Palette->OriginalColor[0] != clBlack)
      {
         TIvoryPalette* Palette = TheIvoryDraw->Palette;
         unsigned char ix0 = (unsigned char)Palette->GetNearestPaletteIndex(Palette->OriginalColor[0]);
         unsigned char ix255 = (unsigned char)Palette->GetNearestPaletteIndex(Palette->OriginalColor[255]);
         for(int i = 0; i < height; i++)
         {
            for(int col = 0; col < width; col++)
            {
               if(*Memory == 0) *offscr = ix0;
               else if(*Memory == 255) *offscr = ix255;
               else *offscr = *Memory;
               Memory++;
               offscr++;
            }
            offscr += Pitch - width;
            Memory -= width + width;
         }
      }
      else
      {
         for(int i = 0; i < height; i++)
         {
            memcpy(offscr, Memory, width);
            offscr += Pitch;
            Memory -= width;
         }
      }
   }
   catch(...)
   {
      throw;
   }
}
//---------------------------------------------------------------------------
void __fastcall TIvoryBitmapSurface::FastLoadPCX(unsigned char* Memory, int width, int height, int offsetpixels, int sizepixels)
{
   try
   {
      Invalidate();
      if(OnChange) OnChange(this);

      int HowMany, Count, CountTmp;
      int sizex = width, sizextmp;
      int sizey = height, sizeytmp, y;
      Memory += offsetpixels;
      unsigned char* outptr;

      sizeytmp = sizey;
      sizextmp = sizex;
      y = 0;
      outptr = (unsigned char*)Pixels;
      unsigned char ix0 = 0;
      unsigned char ix255 = 255;
      if(TheIvoryDraw->Palette->OriginalColor[255] != clWhite ||
         TheIvoryDraw->Palette->OriginalColor[0] != clBlack)
      {
         TIvoryPalette* Palette = TheIvoryDraw->Palette;
         ix0 = (unsigned char)Palette->GetNearestPaletteIndex(Palette->OriginalColor[0]);
         ix255 = (unsigned char)Palette->GetNearestPaletteIndex(Palette->OriginalColor[255]);
      }
      while(sizeytmp)
      {
         unsigned char pixel = *Memory++;
         if((pixel & 0xC0) == 0xC0)
         {
            HowMany = pixel & 0x3F;
            pixel = *Memory++;
            while(HowMany && sizeytmp)
            {
               Count = min(HowMany, sizextmp);
               CountTmp = Count;
               while(CountTmp--)
               {
                  if(pixel == 0) *outptr = ix0;
                  else if(pixel == 255) *outptr = ix255;
                  else *outptr = pixel;
                  outptr++;
               }
               HowMany -= Count;
               sizextmp -= Count;
               if(sizextmp == 0)
               {
                  sizextmp = sizex;
                  sizeytmp--; y++;
                  outptr = (unsigned char*)ScanLine[y];
               }
            }
         }
         else
         {
            if(pixel == 0) *outptr = ix0;
            else if(pixel == 255) *outptr = ix255;
            else *outptr = pixel;
            outptr++;
            sizextmp--;
            if(sizextmp == 0)
            {
               sizextmp = sizex;
               sizeytmp--; y++;
               outptr = (unsigned char*)ScanLine[y];
            }
         }
      }
   }
   catch(...)
   {
      throw;
   }
}
//---------------------------------------------------------------------------
void __fastcall TIvoryBitmapSurface::LoadFile()
{
   const char* err_msg = "Error loading image file";
   const char* ivoryfuncname = "LoadFile";
   int FileHandle = NULL;
   try
   {
      FileHandle = FileOpen(FileName, fmOpenRead);
      if(!FileHandle) IVORYERROR(err_msg, ierOther);
      Size = FileSeek(FileHandle, 0, 2);
      if(Size < sizeof(TPcxHeader)) IVORYERROR(err_msg, ierOther);
      Memory = new unsigned char[Size];
      FileSeek(FileHandle, 0, 0);
      FileRead(FileHandle, Memory, Size);
      FileClose(FileHandle); FileHandle = NULL;
      int offsetpal, sizepal;
      ReadFileInfo(Memory, Size, false, FileType, ImageWidth, ImageHeight, offsetpal, sizepal, OffsetPixels, SizePixels);
      if(FileType > 0)
      {
         if(Size < offsetpal + sizepal) IVORYERROR(err_msg, ierOther);
         if(Size < OffsetPixels + SizePixels) IVORYERROR(err_msg, ierOther);
      }
      else
      {
         // unknown file format
         ImageWidth = 0; ImageHeight = 0;
         if(OnUserFileFormat) OnUserFileFormat(this, Memory, Size, FileType, ImageWidth, ImageHeight);
         if(FileType <= 0 || ImageWidth <= 0 || ImageHeight <= 0) IVORYERROR(err_msg, ierOther);
      }
      Content = 1;
   }
   catch(...)
   {
      if(FileHandle) { FileClose(FileHandle); FileHandle = NULL; }
      if(Memory) { delete [] Memory; Memory = NULL; }
      Content = 0;
      throw;
   }
}
//---------------------------------------------------------------------------
void __fastcall TIvoryBitmapSurface::LoadResource()
{
   const char* err_msg = "Error loading image file";
   const char* ivoryfuncname = "LoadFile";

   try
   {
      bool BitmapResource = true;
      HRSRC hFoundRes = FindResource(DllHandle, ResourceName.c_str(), RT_BITMAP);
      if(!hFoundRes) { hFoundRes = FindResource(DllHandle, ResourceName.c_str(), RT_RCDATA); BitmapResource = false; }
      if(!hFoundRes) IVORYERROR(err_msg, ierOther);
      HGLOBAL hGlobal = ::LoadResource(DllHandle, hFoundRes);
      Memory = (unsigned char*)LockResource(hGlobal);
      Size = SizeofResource(DllHandle, hFoundRes);
      if(!BitmapResource && Size < sizeof(TPcxHeader)) IVORYERROR(err_msg, ierOther);
      int offsetpal, sizepal;
      ReadFileInfo(Memory, Size, BitmapResource, FileType, ImageWidth, ImageHeight, offsetpal, sizepal, OffsetPixels, SizePixels);
      if(FileType > 0)
      {
         if(Size < offsetpal + sizepal) IVORYERROR(err_msg, ierOther);
         if(Size < OffsetPixels + SizePixels) IVORYERROR(err_msg, ierOther);
      }
      else
      {
         // unknown file format
         ImageWidth = 0; ImageHeight = 0;
         if(OnUserFileFormat) OnUserFileFormat(this, Memory, Size, FileType, ImageWidth, ImageHeight);
         if(FileType <= 0 || ImageWidth <= 0 || ImageHeight <= 0) IVORYERROR(err_msg, ierOther);
      }
      Content = 2;
   }
   catch(...)
   {
      Content = 0;
      Memory = NULL;
   }
}
//---------------------------------------------------------------------------
void __fastcall TIvoryBitmapSurface::LoadFromFile()
{
   if(ComponentState.Contains(csDesigning)) return;
   LoadFile();
   if(Content != 0 && TheIvoryDraw != NULL && TheIvoryDraw->DirectDraw != NULL)
   {
      if(Width != ImageWidth || Height != ImageHeight)
         Recreate(ImageWidth, ImageHeight);
      else
         Paint(false);
   }
}
//---------------------------------------------------------------------------
void __fastcall TIvoryBitmapSurface::LoadFromResource()
{
   if(ComponentState.Contains(csDesigning)) return;
   LoadResource();
   if(Content != 0 && TheIvoryDraw != NULL && TheIvoryDraw->DirectDraw != NULL)
   {
      if(Width != ImageWidth || Height != ImageHeight)
         Recreate(ImageWidth, ImageHeight);
      else
         Paint(false);
   }
}
//---------------------------------------------------------------------------

